@@ -53,6 +53,10 @@ void SoftwareRectangleNode::setImage(const QImage &image)
         return;
     }
 
+    if (m_imageNode) {
+        cleanupImageNode();
+    }
+
     m_textureInfo = ShaderNode::TextureInfo{
         .channel = 0,
         .options = {},
@@ -60,6 +64,17 @@ void SoftwareRectangleNode::setImage(const QImage &image)
         .provider = nullptr,
         .providerConnection = {},
     };
+
+    if (!m_textureInfo.texture) {
+        return;
+    }
+
+    m_imageNode = m_window->createImageNode();
+    if (m_imageNode) {
+        m_imageNode->setTexture(m_textureInfo.texture.get());
+        m_imageNode->setFiltering(QSGTexture::Filtering::Linear);
+        appendChildNode(m_imageNode);
+    }
 }
 
 void SoftwareRectangleNode::setTextureProvider(QSGTextureProvider *provider)
@@ -68,6 +83,10 @@ void SoftwareRectangleNode::setTextureProvider(QSGTextureProvider *provider)
         return;
     }
 
+    if (m_imageNode) {
+        cleanupImageNode();
+    }
+
     m_textureInfo = ShaderNode::TextureInfo{
         .channel = 0,
         .options = {},
@@ -75,6 +94,8 @@ void SoftwareRectangleNode::setTextureProvider(QSGTextureProvider *provider)
         .provider = provider,
         .providerConnection = {},
     };
+
+    // The render node will be created in preprocess().
 }
 
 void SoftwareRectangleNode::setRadius(qreal radius)
@@ -114,33 +135,25 @@ QSGRenderNode::RenderingFlags SoftwareRectangleNode::flags() const
 
 void SoftwareRectangleNode::preprocess()
 {
-    QSGTexture *texture = nullptr;
-    if (QSGTextureProvider *provider = m_textureInfo.provider) {
-        texture = provider->texture();
-        if (QSGDynamicTexture *dynamic_texture = qobject_cast<QSGDynamicTexture *>(provider->texture())) {
+    auto provider = m_textureInfo.provider;
+    if (provider) {
+        QSGTexture *texture = provider->texture();
+        if (QSGDynamicTexture *dynamic_texture = qobject_cast<QSGDynamicTexture *>(texture)) {
             dynamic_texture->updateTexture();
         }
-    } else if (m_textureInfo.texture) {
-        texture = m_textureInfo.texture.get();
-    }
 
-    if (texture) {
-        if (!m_imageNode) {
-            m_imageNode = m_window->createImageNode();
-            m_imageNode->setFlag(OwnedByParent);
+        if (texture) {
+            if (!m_imageNode) {
+                m_imageNode = m_window->createImageNode();
+                m_imageNode->setTexture(texture);
+                m_imageNode->setFiltering(QSGTexture::Filtering::Linear);
+                appendChildNode(m_imageNode);
+            } else {
+                m_imageNode->setTexture(texture);
+            }
+        } else if (m_imageNode) {
+            cleanupImageNode();
         }
-
-        // The rect will be set in render().
-        m_imageNode->setFiltering(QSGTexture::Filtering::Linear);
-        m_imageNode->setTexture(texture);
-
-        if (!m_imageNode->parent()) {
-            appendChildNode(m_imageNode);
-        }
-    } else if (m_imageNode) {
-        removeChildNode(m_imageNode);
-        delete m_imageNode;
-        m_imageNode = nullptr;
     }
 }
 
@@ -178,3 +191,10 @@ void SoftwareRectangleNode::render(const RenderState *state)
         m_imageNode->setRect(withoutCorners);
     }
 }
+
+void SoftwareRectangleNode::cleanupImageNode()
+{
+    removeChildNode(m_imageNode);
+    delete m_imageNode;
+    m_imageNode = nullptr;
+}
